/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2002  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 1999-2002  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include "../xmms/i18n.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <dirent.h>
#include <unistd.h>

#include <string.h>

#include <gtk/gtk.h>
#include <stdio.h>

/* XPM */
static char *folder[] = {
  "16 16 16 1",
  " 	c None",
  ".	c #f4f7e4",
  "X	c #dee4b5",
  "o	c #e1e7b9",
  "O	c #c6cba4",
  "+	c #dce2b8",
  "@	c #e9e9ec",
  "#	c #d3d8ae",
  "$	c #d8daca",
  "%	c #b2b2b5",
  "&	c #767862",
  "*	c #e3e6c3",
  "=	c #1b1b1a",
  "-	c #939684",
  ";	c #555555",
  ":	c #000000",
  "                ",
  "                ",
  "  ::::          ",
  " :.@@O:         ",
  ":-&&&&&:::::    ",
  ":.@@@@@*$O#O=   ",
  ":@*+XXXX+##O:   ",
  ":.*#oooXXXXX:   ",
  ":@+XoXXXXXX#:   ",
  ":@*ooXXXXXX#:   ",
  ":@**XXXXXXX#:   ",
  ":@*XXXXXXXX%:   ",
  ":$.*OOOOOO%-:   ",
  " ;:::::::::::   ",
  "                ",
  "                "
};

/* Icon by Jakub Steiner <jimmac@ximian.com> */

/* XPM */
static char *ofolder[] = {
  "16 16 16 1",
  " 	c None",
  ".	c #a9ad93",
  "X	c #60634d",
  "o	c #dee4b5",
  "O	c #9ca085",
  "+	c #0c0d04",
  "@	c #2f2f31",
  "#	c #3b3d2c",
  "$	c #c8cda2",
  "%	c #e6e6e9",
  "&	c #b3b5a5",
  "*	c #80826d",
  "=	c #292a1c",
  "-	c #fefef6",
  ";	c #8f937b",
  ":	c #000000",
  "                ",
  "                ",
  "  ::::          ",
  " :-%%&:         ",
  ":-;;;OX:::::    ",
  ":-;;;;O;O;&.:   ",
  ":-*X##@@@@@=#:  ",
  ":%*+-%%ooooooO: ",
  ":%X;%ooooooo.*: ",
  ":.+-%oooooooO:  ",
  ":*O-oooooooo*:  ",
  ":O-oooooooo.:   ",
  ":*-%$$$$$$OX:   ",
  " :::::::::::    ",
  "                ",
  "                "
};

#define NODE_SPACING 4

static GdkPixbuf *folder_pixmap = NULL, *ofolder_pixmap;
GtkTreeViewColumn *column;

gboolean
get_first_dir_entry (gchar * path, GDir ** dir, gchar ** file)
{
  *dir = g_dir_open (path, 0, NULL);

  if (!*dir)
    {
      return FALSE;
    }

  *file = (gchar *) g_dir_read_name (*dir);

  if (!file)
    {
      g_dir_close (*dir);
      return FALSE;
    }
  return TRUE;
}

gchar *
get_next_dir_entry (GDir ** dir)
{
  return (gchar *) g_dir_read_name (*dir);
}

gchar *
get_full_path (GtkTreeModel * model, GtkTreeIter * iter)
{
  gchar *tmp;
  GSList *list = NULL, *tmp2 = NULL, *tmp3 = NULL;
  GtkTreeIter iter2;
  GtkTreeIter *itr = gtk_tree_iter_copy (iter);

  gtk_tree_model_get (model, itr, 1, &tmp, -1);
  if (tmp)
    {
      list = g_slist_prepend (list, g_strdup (tmp));
      g_free (tmp);
    }

  tmp = NULL;

  while (gtk_tree_model_iter_parent (model, &iter2, itr))
    {
      gtk_tree_model_get (model, &iter2, 1, &tmp, -1);
      if (tmp)
	{
	  list = g_slist_prepend (list, g_strdup (tmp));
	  g_free (tmp);
	  tmp = NULL;
	}
      gtk_tree_iter_free (itr);
      itr = gtk_tree_iter_copy (&iter2);
    }
  tmp = NULL;

  tmp2 = list;
  while (tmp2)
    {
      if (tmp)
	{
	  gint x = strlen (tmp);
	  tmp = g_realloc (tmp, x + strlen (tmp2->data) + 2);
	  tmp[x] = '/';
	  tmp[x + 1] = '\0';
	}
      else
	{
	  gint x = strlen (tmp2->data);
	  tmp = g_realloc (tmp, x + 2);
	  tmp[0] = '/';
	  tmp[1] = '\0';
	}
      strcat (tmp, tmp2->data);
      tmp3 = g_slist_next (tmp2);
      tmp2 = tmp3;
    }
  tmp[strlen (tmp)] = '\0';

  g_slist_foreach (list, (GFunc) g_free, NULL);
  g_slist_free (list);
  return tmp;
}

void
row_expanded_cb (GtkTreeView * treeview,
		 GtkTreeIter * arg1, GtkTreePath * arg2, gpointer user_data)
{
  GtkTreeIter child_iter, iter;
  GtkTreeModel *model;
  gchar *name, *tmp;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  gtk_tree_model_get (GTK_TREE_MODEL (model), arg1, 1, &name, -1);

  gtk_tree_store_set (GTK_TREE_STORE (model), arg1, 0, ofolder_pixmap, -1);

  if (name)
    {
      GDir *dir;

      gchar *full_path = get_full_path (model, arg1);

      if (!full_path)
	{
	  return;
	}

      gtk_tree_model_iter_children (GTK_TREE_MODEL (model),
				    &child_iter, arg1);

      if (!get_first_dir_entry (full_path, &dir, &name))
	{
	  g_free (full_path);
	  return;
	}

      tmp = g_strdup_printf ("%s/%s", full_path, name);
      if (!g_file_test (tmp, G_FILE_TEST_IS_DIR))
	{
	  g_free (tmp);
	  while ((name = get_next_dir_entry (&dir)) != NULL)
	    {
	      tmp = g_strdup_printf ("%s/%s", full_path, name);
	      if (!g_file_test (tmp, G_FILE_TEST_IS_DIR))
		{
		  g_free (tmp);
		  continue;
		}
	      else
		{
		  gtk_tree_store_set (GTK_TREE_STORE (model), &child_iter, 0,
				      folder_pixmap, 1, name, -1);
		  gtk_tree_store_append (GTK_TREE_STORE (model), &iter,
					 &child_iter);
		}
	      break;
	    }
	}
      else
	{
	  gtk_tree_store_set (GTK_TREE_STORE (model), &child_iter, 0,
			      folder_pixmap, 1, name, -1);
	  gtk_tree_store_append (GTK_TREE_STORE (model), &iter, &child_iter);
	}
      g_free (tmp);

      while ((name = get_next_dir_entry (&dir)) != NULL)
	{
	  tmp = g_strdup_printf ("%s/%s", full_path, name);
	  if (g_file_test (tmp, G_FILE_TEST_IS_DIR))
	    {
	      gtk_tree_store_append (GTK_TREE_STORE (model), &child_iter,
				     arg1);
	      gtk_tree_store_set (GTK_TREE_STORE (model), &child_iter, 0,
				  folder_pixmap, 1, name, -1);
	      gtk_tree_store_append (GTK_TREE_STORE (model), &iter,
				     &child_iter);
	    }
	  g_free (tmp);
	}
      g_dir_close (dir);
      g_free (full_path);
    }
  else
    {
      return;
    }
}

static void
row_collapsed_cb (GtkTreeView * tree_view,
		  GtkTreeIter * iter, GtkTreePath * path)
{
  GtkTreeIter child_iter;
  GtkTreeModel *model;

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree_view));
  gtk_tree_store_set (GTK_TREE_STORE (model), iter, 0, folder_pixmap, -1);

  gtk_tree_model_iter_children (GTK_TREE_MODEL (model), &child_iter, iter);
  while (gtk_tree_store_remove (GTK_TREE_STORE (model), &child_iter))
    {
      ;
    }
  gtk_tree_store_append (GTK_TREE_STORE (model), &child_iter, iter);
}

static void
row_activated_cb (GtkTreeView * treeview,
		  GtkTreePath * arg1,
		  GtkTreeViewColumn * arg2, gpointer user_data)
{
  gchar *path;
  GtkTreeIter iter;
  GtkTreeModel *model;
  void (*handler) (gchar *);

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_model_get_iter (model, &iter, arg1))
    {
      return;
    }

  path = get_full_path (model, &iter);
  if (!path)
    {
      return;
    }

  if (!gtk_tree_model_get_iter (model, &iter, arg1))
    {
      g_free (path);
      return;
    }

  handler =
    (void (*)(gchar *)) g_object_get_data (G_OBJECT (treeview), "user_data");

  if (handler)
    {
      handler (path);
    }

  g_free (path);
}

static void
show_cb (GtkWidget * widget, gpointer data)
{
  GtkTreePath *path;
  GtkWidget *tv = GTK_WIDGET (data);
  path = gtk_tree_path_new_first ();
  gtk_tree_view_expand_row (GTK_TREE_VIEW (tv), path, FALSE);
  gtk_tree_path_free (path);

/*	GtkCTree *tree = GTK_CTREE(data);
	GtkCTreeNode *node = gtk_object_get_data(GTK_OBJECT(tree),
						 "selected_node");

	if (node)
		gtk_ctree_node_moveto(tree, node, -1, 0.6, 0);
*/
}

static void
ok_clicked (GtkWidget * widget, GtkWidget * tree)
{
  GtkWidget *window;
  GtkTreeSelection *selection;
  GtkTreeIter iter;
  GtkTreeModel *model;
  gchar *path = NULL;
  void (*handler) (gchar *);

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tree));
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));

  window = g_object_get_data (G_OBJECT (widget), "user_data");
  gtk_widget_hide (window);

  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      path = get_full_path (model, &iter);
      handler =
	(void (*)(gchar *)) g_object_get_data (G_OBJECT (tree), "user_data");

      if (handler)
	{
	  handler (path);
	}
    }
  gtk_widget_destroy (window);
  g_free (path);
}

static void
init_tree (GtkTreeModel * model, gchar * path)
{
  GtkTreeIter piter, iter;
  gtk_tree_store_append (GTK_TREE_STORE (model), &piter, NULL);
  gtk_tree_store_set (GTK_TREE_STORE (model), &piter, 0, folder_pixmap, 1,
		      path, -1);
  gtk_tree_store_append (GTK_TREE_STORE (model), &iter, &piter);
}

GtkWidget *
xmms_create_dir_browser (char *title, char *current_path,
			 GtkSelectionMode mode, void (*handler) (char *))
{
  GtkWidget *window, *scroll_win, *vbox, *bbox, *ok, *cancel, *sep, *tv;
  GtkTreeModel *model;
  GtkTreeSelection *selection;

  window = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_widget_set_size_request (GTK_WIDGET (window), 300, 400);
  gtk_container_set_border_width (GTK_CONTAINER (window), 10);

  vbox = gtk_vbox_new (FALSE, 10);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (window)->vbox), vbox);

  scroll_win = gtk_scrolled_window_new (NULL, NULL);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll_win),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  gtk_box_pack_start (GTK_BOX (vbox), scroll_win, TRUE, TRUE, 0);
  gtk_widget_show (scroll_win);

  gtk_widget_realize (window);
  if (!folder_pixmap)
    {
      folder_pixmap = gdk_pixbuf_new_from_xpm_data ((const gchar **) folder);
      ofolder_pixmap =
	gdk_pixbuf_new_from_xpm_data ((const gchar **) ofolder);
    }

  (GtkTreeStore *) model =
    gtk_tree_store_new (2, GDK_TYPE_PIXBUF, G_TYPE_STRING);

  tv = gtk_tree_view_new_with_model (GTK_TREE_MODEL (model));
  gtk_widget_show (tv);
  g_object_unref (model);
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tv));

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tv));
  gtk_tree_selection_set_mode (selection, mode);

  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (tv), FALSE);

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (tv));

  column = gtk_tree_view_column_new_with_attributes ("Icon",
						     gtk_cell_renderer_pixbuf_new
						     (), "pixbuf", 0, NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
  gtk_tree_view_append_column (GTK_TREE_VIEW (tv), column);

  column = gtk_tree_view_column_new_with_attributes ("Tree",
						     gtk_cell_renderer_text_new
						     (), "text", 1, NULL);
  gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
  gtk_tree_view_append_column (GTK_TREE_VIEW (tv), column);

  gtk_container_add (GTK_CONTAINER (scroll_win), tv);

  g_signal_connect_after (G_OBJECT (tv), "row-expanded",
			  G_CALLBACK (row_expanded_cb), NULL);

  g_signal_connect_after (G_OBJECT (tv), "row_collapsed",
			  G_CALLBACK (row_collapsed_cb), NULL);

  g_signal_connect (G_OBJECT (tv), "row_activated",
		    G_CALLBACK (row_activated_cb), NULL);
  g_signal_connect (GTK_OBJECT (window), "show",
		    GTK_SIGNAL_FUNC (show_cb), tv);

  g_object_set_data (G_OBJECT (tv), "user_data", (gpointer) handler);

  sep = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, FALSE, 0);
  gtk_widget_show (sep);

  bbox = gtk_hbutton_box_new ();
  gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (bbox), 5);

  ok = gtk_button_new_from_stock (GTK_STOCK_OK);
  g_object_set_data (G_OBJECT (ok), "user_data", window);

  GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
  gtk_window_set_default (GTK_WINDOW (window), ok);
  gtk_box_pack_start (GTK_BOX (bbox), ok, TRUE, TRUE, 0);
  g_signal_connect (G_OBJECT (ok), "clicked", G_CALLBACK (ok_clicked), tv);
  gtk_widget_show (ok);

  cancel = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
  GTK_WIDGET_SET_FLAGS (cancel, GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (bbox), cancel, TRUE, TRUE, 0);
  g_signal_connect_swapped (G_OBJECT (cancel), "clicked",
			    G_CALLBACK (gtk_widget_destroy), window);
  gtk_widget_show (cancel);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), bbox, FALSE,
		      FALSE, 0);
  gtk_widget_show (bbox);
  gtk_widget_show (vbox);

  init_tree (model, "/");

  if (current_path && *current_path)
    {
      char **dir;
      int i;
      GtkTreeIter iter;
      printf ("%s\n\n", current_path);
      dir = g_strsplit (current_path, "/", 0);
      gtk_tree_model_get_iter_first (GTK_TREE_MODEL (model), &iter);

      for (i = 1; dir[i] != NULL; i++)
	{
	  gchar *name = NULL;
printf("%s\n", dir[i]);
	  if (dir[i][0] == '\0')
	    {
printf("continue\n");
	      continue;
	    }
	  gtk_tree_model_get (model, &iter, 1, &name, -1);
	  if (!name)
	    {
printf("break\n");
	      break;
	    }
printf("%s %s\n" , name, dir[i]);
	  if (!strcmp (name, dir[i]))
	    {
	      if ((dir[i + 1]) && (gtk_tree_model_iter_next (model, &iter)))
		{
		  GtkTreePath *tpath = gtk_tree_model_get_path (model, &iter);
		  gtk_tree_view_expand_row (GTK_TREE_VIEW (tv), tpath, FALSE);
		  gtk_tree_path_free (tpath);
printf("expand\n");
		}
	      else
		{
		  GtkTreePath *tpath = gtk_tree_model_get_path (model, &iter);
		  gtk_tree_selection_select_iter (gtk_tree_view_get_selection
						  (GTK_TREE_VIEW (tv)),
						  &iter);
		  gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (tv), tpath,
						NULL, FALSE, 0, 0);
		  gtk_tree_path_free (tpath);
printf("select\n");
		}
	    }
	  g_free (name);
	}
      g_strfreev (dir);
    }

/*	gtk_ctree_select(ctree, selected_node);
	gtk_object_set_data(GTK_OBJECT(tree), "selected_node", selected_node);
*/
  return window;
}
